from gantry.logger.main import (
    get_client,
    init,
    instrument,
    log_feedback,
    log_feedback_event,
    log_file,
    log_prediction_event,
    log_predictions,
    log_record,
    log_records,
    ping,
    ready,
    setup_logger,
)
from gantry.utils import compute_feedback_id

__all__ = [
    "init",
    "instrument",
    "log_record",
    "log_records",
    "log_file",
    "ping",
    "ready",
    "log_feedback",
    "log_predictions",
    "log_feedback_event",
    "log_prediction_event",
    "get_client",
    "setup_logger",
    "compute_feedback_id",
]



import json
from io import StringIO

from gantry.logger.utils import _build_batch_iterator


def test_build_batch_iterator():
    event_list = [{1: 1}, {2: 2}, {3: 3}, {4: 4}, {5: 5}]
    # We should only have 3 batches created from the list.
    batch_iter = _build_batch_iterator(event_list, 2)
    iters = 0
    for _ in batch_iter:
        iters += 1
    assert iters == 3

    # We should still have all 5 lines in the file
    batch_iter = _build_batch_iterator(event_list, 2)
    result = "".join([part.decode("utf-8") for part in batch_iter])
    file = StringIO(result)

    for line in file.readlines():
        json.loads(line)

    file.seek(0)
    assert len(file.readlines()) == 5



import collections
import collections.abc
import datetime
import hashlib
import json
import logging
import re
import sys
from typing import Any, Dict, List, Optional, Tuple

from dateutil import parser

from gantry.serializers import EventEncoder

logger = logging.getLogger(__name__)


def to_datetime(s: str) -> datetime.datetime:
    """
    Converts a string to a naive UTC datetime object
    """
    try:
        # fromisoformat is present starting in python3.7
        if s.endswith("Z"):
            dt = datetime.datetime.fromisoformat(s.replace("Z", "+00:00"))
        else:
            dt = datetime.datetime.fromisoformat(s)
    except (AttributeError, ValueError, TypeError):
        try:
            dt = datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%f")
        except Exception:
            dt = parser.parse(s)
    if _is_offset_naive(dt):
        return dt

    naive_dt = dt.astimezone(datetime.timezone.utc).replace(tzinfo=None)
    return naive_dt


def to_isoformat_duration(relative_interval: datetime.timedelta) -> str:
    days = relative_interval.days
    seconds = relative_interval.seconds

    minutes, seconds = seconds // 60, seconds % 60
    hours, minutes = minutes // 60, minutes % 60
    weeks, days = days // 7, days % 7

    timedelta_str = "P"
    if weeks > 0:
        timedelta_str += str(weeks) + "W"
    if days > 0:
        timedelta_str += str(days) + "D"
    timedelta_str += "T"
    if hours > 0:
        timedelta_str += str(hours) + "H"
    if minutes > 0:
        timedelta_str += str(minutes) + "M"
    if seconds > 0:
        timedelta_str += str(seconds) + "S"

    if timedelta_str[-1] == "T":
        timedelta_str = timedelta_str[:-1]

    if len(timedelta_str) == 1:
        return "P0D"

    return timedelta_str


def to_timestamp(dt) -> float:
    """
    converts naive Datetime object to a timestamp in seconds.

    implemented for py 2 compatibility:
    https://stackoverflow.com/questions/8777753/converting-datetime-date-to-utc-timestamp-in-python
    """
    return (dt - datetime.datetime(1970, 1, 1)).total_seconds()


def _is_offset_naive(d: datetime.datetime) -> bool:
    # https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
    aware = d.tzinfo is not None and d.tzinfo.utcoffset(d) is not None
    return not aware


def check_event_time_in_future(event_timestamp: datetime.datetime) -> bool:
    """
    Raises an error if the event timestamp is in the future.
    Note, `event_timestamp` is assumed to be in UTC. This
    method will handle both offset-naive and offset-aware event_timestamp
    """
    current_time = (
        datetime.datetime.utcnow()
        if _is_offset_naive(event_timestamp)
        else datetime.datetime.now(datetime.timezone.utc)
    )
    return event_timestamp > current_time


def compute_feedback_id(inputs: Dict[str, Any], feedback_keys: Optional[List[str]] = None) -> str:
    assert isinstance(inputs, collections.abc.Mapping)
    if not feedback_keys:
        # When not specified, default to feedback_keys are all the
        # fields of inputs
        feedback_keys = list(inputs.keys())

    values = []
    for key in sorted(feedback_keys):
        input_value = inputs[key]
        # TODO change to hashable type if input_value isn't
        # hashable
        values.append(input_value)

    return hashlib.md5(
        json.dumps(values, sort_keys=True, cls=EventEncoder).encode("utf-8")
    ).hexdigest()


def clean_name(name: str) -> str:
    """Replace any non-alphanumeric characters, except '-' and '.', with '-'"""
    return re.sub(r"[^a-zA-Z0-9\-.]+", "-", name)


def generate_gantry_name(name: str, max_len: int = 64) -> str:
    prefix = "gantry-"
    suffix = "-" + hashlib.sha256(name.encode("utf-8")).hexdigest()[:8]

    remaining_len = max_len - len(prefix) - len(suffix)
    trunc_name = clean_name(name)[:remaining_len].lower().strip("-")

    return prefix + trunc_name + suffix


def parse_s3_path(path: str) -> Tuple[str, str]:
    """Parses an S3 path of the form s3://bucket/path/to/obj and returns (key, path/to/obj)"""
    prefix_len = len("s3://")
    s3_path = path[prefix_len:]
    bucket, _, key = s3_path.partition("/")

    return (bucket, key)


def format_msg_with_color(msg: str, color: str, logger: logging.Logger) -> str:
    """Given a message and a colorama color, conditionally returns a formatted message if the only
    handlers of the gantry package logger are NullHandlers or StreamHandlers with sys.stderr
    or sys.stdout
    """
    # loop to short-circuit and return original message if a potential file handler is found
    for handler in logging.getLogger("gantry").handlers:
        if isinstance(handler, logging.NullHandler):
            # It is okay if handler is NullHandler
            continue
        elif isinstance(handler, logging.StreamHandler):
            # If handler is a StreamHandler make sure stream is sys.stderr or sys.stdout
            if handler.stream in (
                sys.stderr,
                sys.stdout,
            ):
                continue
        else:
            # In all other cases, we might be logging to a file so just return the original msg
            return msg
    # If we haven't short-circuited and returned original message, return the colored message
    return color + msg


def obj_by_id(els):
    obj = {}
    for el in els:
        obj[el.id] = el

    return obj



import uuid
from typing import Optional


class GantryException(Exception):
    """
    Base class for Gantry exceptions.
    """

    pass


class GantryLoggingException(GantryException):
    """
    Raised when logging invalid data.
    """

    pass


class GantryLoggingDataTypeError(GantryLoggingException):
    """
    Raised when Gantry cannot process datatype
    """

    pass


class ConfigurationMissing(GantryException):
    """
    There is some configuration missing
    """

    def __init__(self, message=""):
        if len(message) == 0:
            message = "Gantry is missing some configuration"
        else:
            message = "Gantry is missing some configuration: {}".format(message)

        super(ConfigurationMissing, self).__init__(message)


class InvalidConfigError(GantryException):
    """
    Invalid configuration passed in
    """

    pass


class ClientNotInitialized(GantryException):
    """
    Gantry client not initialized
    """

    def __init__(self):
        message = "Gantry client is not initialized. Did you call `gantry.init()`?"
        super(ClientNotInitialized, self).__init__(message)


class DatabaseAccessError(GantryException):
    """
    Gantry client not initialized
    """

    def __init__(self, message=""):
        if len(message) == 0:
            message = "Unable to access database, check the configuration"
        super(DatabaseAccessError, self).__init__(message)


class FeedbackEventMissingPrediction(GantryException):
    """
    Feedback event cannot be processed right now because we cannot find the
    corresponding predictions. Could be raised because the prediction has not
    been ingested or processed by the system.
    """

    def __init__(self, fe_id):
        message = "Feedback event (id={}) is missing prediction.".format(fe_id)
        super(FeedbackEventMissingPrediction, self).__init__(message)


class MissingDependencyError(GantryException):
    """
    Raised when a dependency needed for a feature is not installed.
    """

    pass


class FeatureTypeInvalid(GantryException):
    def __init__(self, msg):
        super(FeatureTypeInvalid, self).__init__(msg)


class InvalidAttributeError(GantryException):
    """
    Raised when the attribute requested is invalid for a sketch.
    """

    def __init__(self, msg):
        super(InvalidAttributeError, self).__init__(msg)


class GantryAPIException(GantryException):
    """
    Base class for exceptions in the API
    """

    status_code = 400

    # Based on:
    # https://flask.palletsprojects.com/en/2.0.x/errorhandling/#returning-api-errors-as-json
    def __init__(self, msg, status_code: Optional[int] = None):
        """Optional status_code can be passed and it will be used
        by flask as HTTP response status code.
        """
        super().__init__(msg)
        if status_code is not None:
            self.status_code = status_code


class MissingQueryError(GantryAPIException):
    """
    Raised when a query is missing
    """

    pass


class QueryError(GantryAPIException):
    """
    Raised when a query doesn't have a result
    """

    pass


class InvalidVersionError(GantryAPIException):
    """
    Raised when the function version in the request is invalid
    """

    pass


class AuthenticationError(GantryAPIException):
    """
    Raised when the request does not have valid authentication
    """

    status_code = 401


class AccessDenied(GantryAPIException):
    """
    Raised when the user does not have access to the requested resource
    """

    status_code = 404


class GantryFilterException(GantryException):
    """
    Raised by the sdk when there is an exception to a filter
    """

    pass


class GantryDockerException(GantryException):
    """
    Raised by the Gantry Docker command when docker or docker-compose is not found
    """

    pass


class GantryDatabaseConnectionError(GantryException):
    """
    Raised when Gantry cannot connect to the database
    """

    pass


class ApplicationNotFoundException(GantryException):
    """Raised when an application is not found"""

    def __init__(
        self,
        application_node_id: Optional[uuid.UUID] = None,
        func_name: Optional[str] = None,
        version: Optional[int] = None,
    ):
        if application_node_id:
            message = "Could not find an application with application id: {}".format(
                application_node_id
            )
        elif func_name and version:
            message = (
                "Could not find an application with application name: {} and version: {}".format(
                    func_name, version
                )
            )
        else:
            message = "Unable to find that application"

        super(GantryException, self).__init__(message)


class ApplicationDisabledException(GantryException):
    """Raised when an application is disabled and the action cannot be performed"""

    pass


class GantryBatchCreationException(GantryException):
    """Raised when gantry failed to create batch"""

    pass
